#include <qcolor.h>
#include <qevent.h>
#include <qarray.h>
#include <qimage.h>

#include "../../../Core/Configuration/src/Configuration.h"
#include "../../../Utilities/src/Matrix.h"
#include "../../../Core/Interfaces/src/RendererInterface.h"
#include "../../../Core/Interfaces/src/FrameBufferInterface.h"
#include "HGLSplatRendererConfiguration.h"
#include "HGLSplatRenderer.h"

// **************
// public methods
// **************

HGLSplatRenderer::HGLSplatRenderer()
  : rendererWidget(0)
{
}

HGLSplatRenderer::~HGLSplatRenderer()
{
  if (rendererWidget != 0) {
    delete rendererWidget;
  }
}

void HGLSplatRenderer::initialize (const bool isVisible, QWidget *parent)
{
  // delete the old widget, if necessary
  if (rendererWidget != 0) {
    delete rendererWidget;
  }

  // we allocate the renderer widget in any case, since the HGLSplat rendering is implemented there
  rendererWidget = new HGLSplatRendererWidget (isVisible, &rendererConfiguration, parent);
  // get informed whenever the rendererWidget is repainted
  this->connect (rendererWidget, SIGNAL (repainted()),
		 this, SLOT (handleRepainted()));
}

void HGLSplatRenderer::setVisible (const bool enable)
{
  rendererWidget->setVisible (enable);
}

bool HGLSplatRenderer::isVisible() const
{
  return rendererWidget->isVisible();
}

bool HGLSplatRenderer::isInteractive() const
{
  return false;
}


bool HGLSplatRenderer::providesPointPicking () const
{
  return false;
}

float HGLSplatRenderer::getDepth (unsigned int x, unsigned int y) const
{
  (void)x; (void)y;
  return 0.0f;
}


Vector3D HGLSplatRenderer::getCameraSpaceNormal (unsigned int x, unsigned int y) const
{
  (void)x; (void)y;
  return Vector3D (0, 0, 1);
}



QWidget *HGLSplatRenderer::getRendererWidget() const
{
  return rendererWidget;
}

RendererConfigurationInterface *HGLSplatRenderer::getRendererConfiguration()
{
  return &rendererConfiguration;
}

QImage HGLSplatRenderer::getRenderedImage()
{
  QImage image;

  image = rendererWidget->convertFrameBufferToQImage();
  return image;
}

void HGLSplatRenderer::setSceneView (const MyDataTypes::TransformationMatrix16f newSceneViewMatrix)
{
  rendererWidget->setSceneView (newSceneViewMatrix);
}

void HGLSplatRenderer::getSceneView (MyDataTypes::TransformationMatrix16f sceneViewMatrix) const
{
  rendererWidget->getSceneView (sceneViewMatrix);
}

void HGLSplatRenderer::setViewFrustum (const MyDataTypes::ViewFrustum newViewFrustum)
{
  rendererWidget->setViewFrustum (newViewFrustum);
}

MyDataTypes::ViewFrustum HGLSplatRenderer::getViewFrustum() const
{
  return rendererWidget->getViewFrustum();
}

void HGLSplatRenderer::setViewPortSize (const QSize newViewPortSize)
{
  rendererWidget->setViewPortSize (newViewPortSize);
}

QSize HGLSplatRenderer::getViewPortSize() const
{
  return rendererWidget->getViewPortSize();
}

void HGLSplatRenderer::setShadingEnabled (const bool enable) {
  rendererWidget->setShadingEnabled (enable);
}

bool HGLSplatRenderer::isShadingEnabled() const
{
  return rendererWidget->isShadingEnabled();
}

void HGLSplatRenderer::setLightDirection (const Vector3D newLightDirection)
{
  rendererWidget->setLightDirection (newLightDirection);
}

Vector3D HGLSplatRenderer::getLightDirection() const
{
  return rendererWidget->getLightDirection();
}

void HGLSplatRenderer::setBackgroundColor(QRgb newBackgroundColor) {
	rendererWidget->setBackgroundColor(newBackgroundColor);
}

QRgb HGLSplatRenderer::getBackgroundColor() const {
	return rendererWidget->getBackgroundColor();
}

void HGLSplatRenderer::setTwoSidedNormalsEnabled(const bool enable)
{
  (void)enable;
}

bool HGLSplatRenderer::isTwoSidedNormalsEnabled() const
{
  return true;
}

void HGLSplatRenderer::setShadowsEnabled(const bool enable)
{
  (void)enable;
}

bool HGLSplatRenderer::isShadowsEnabled() const
{
  return false;
}


void HGLSplatRenderer::toggleSelectionVisualization() {
	rendererWidget->setSelectionVisualizationEnabled(!rendererWidget->isSelectionVisualizationEnabled());
}

void HGLSplatRenderer::setSelectionVisualizationEnabled (const bool enable)
{
  rendererWidget->setSelectionVisualizationEnabled(enable);
}

bool HGLSplatRenderer::isSelectionVisualizationEnabled() const
{
  return rendererWidget->isSelectionVisualizationEnabled();
}

void HGLSplatRenderer::setCutoffRadius (const float radius)
{
  rendererConfiguration.setCutOffRadius (radius);
}

float HGLSplatRenderer::getCutoffRadius() const
{
  return rendererConfiguration.getCutOffRadius();
}

void HGLSplatRenderer::setBlendingThresholds (const float constThreshold,
					    const float distThreshold,
					    const float angleThreshold)
{
  (void)constThreshold; (void)distThreshold; (void)angleThreshold;
}

void HGLSplatRenderer::getBlendingThresholds (float *constThreshold, float *distThreshold, float *angleThreshold)
{
  // return default values
  *constThreshold = 1.0f;
  *distThreshold  = 1.0f;
  *angleThreshold = 1.0f;
}

void HGLSplatRenderer::renderFrame (const int, const bool)
{
	rendererWidget->repaint();
}

void HGLSplatRenderer::addToFrame (const std::vector<SurfelInterface *>* surfels, int nSurfels,
				 const int attributes, const bool updateRendererWidget)
{
  (void)surfels; (void)nSurfels; (void)attributes; (void)updateRendererWidget;
}

void HGLSplatRenderer::subtractFromFrame (const std::vector<SurfelInterface *>* surfels, int nSurfels,
					const int attributes, const bool updateRendererWidget)
{
  (void)surfels; (void)nSurfels; (void)attributes; (void)updateRendererWidget;
}

FrameBufferInterface *HGLSplatRenderer::getFrameBuffer() const
{
  // return dummy value (not supported)
  return 0;
}
  

///////////////////////////////////////////////////////////////////////////////
    
// *************
// private slots
// *************

void HGLSplatRenderer::handleRepainted()
{
  emit rendererWidgetRepainted();
}

// ****************
// access functions
// ****************

RendererInterface *createRenderer()
{
  return new HGLSplatRenderer();
}

const char *getRendererType()
{
  return "HGLSplat (Preview)";
}
