#include <qcolor.h>
#include <qevent.h>
#include <qarray.h>
#include <qimage.h>
#include <qpainter.h>
#include <qtimer.h>

#include <crtdbg.h>

#if defined(WIN32)
#  include <windows.h>
#  pragma warning (disable : 4786)
#elif defined(UNIX)
#  include <GL/glx.h>
#endif
#define GLH_EXT_SINGLE_FILE
#include <glh/glh_extensions.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/Configuration/src/Configuration.h"
#include "../../../Core/src/DefaultMouseHandler.h"
#include "../../../Utilities/src/Matrix.h"
#include "../../../DataTypes/src/MyDataTypes.h"
#include "HGLSplatRendererConfiguration.h"
#include "HGLSplatRendererWidget.h"


// **************
// public methods
// **************

HGLSplatRendererWidget::HGLSplatRendererWidget (const bool isVisible, HGLSplatRendererConfiguration *rendererConfiguration,
											QWidget *parent, const char *name)
  : QGLWidget (QGLFormat(AlphaChannel), parent, name), gaussRes(256)
{
	Scene         *scene;
	Configuration *configuration;

	this->rendererConfiguration = rendererConfiguration;


	// disable clearing the background before each paint event
	this->setBackgroundMode (NoBackground);

	visible = isVisible;
	updateViewPortSize = true;
	displayListsValid = false;
	shadingEnabled = true;
	blendingEnabled = true;

	configuration = Configuration::getInstance();

	scene = Scene::getInstance();

	// get informed whenever the Objects in the Scene change
	this->connect (scene, SIGNAL (objectModified(bool)),
				 this, SLOT (handleObjectModified()));

	// get informed whenever the active Object in the Scene changes
	this->connect (scene, SIGNAL (activeObjectChanged()),
				 this, SLOT (handleActiveObjectChanged()));

	// get informed whenever Objects are added to the Scene
	this->connect (scene, SIGNAL (objectAdded()),
				 this, SLOT (handleObjectAdded()));

	// get informed whenever Objects are added to the Scene
	this->connect (scene, SIGNAL (objectRemoved()),
				 this, SLOT (handleObjectRemoved()));

	this->connect (rendererConfiguration, SIGNAL (cutOffRadiusChanged (const float)),
				 this, SLOT (handleCutOffRadiusChanged (const float)));

	this->connect (rendererConfiguration, SIGNAL (conditionThresholdChanged (const float)),
				 this, SLOT (handleConditionThresholdChanged (const float)));

	this->connect (rendererConfiguration, SIGNAL (depthOffsetChanged (const float)),
				 this, SLOT (handleDepthOffsetChanged (const float)));

	// get informed whenever all the settings in the HGLSplatRendererConfiguration have been applied
	this->connect (rendererConfiguration, SIGNAL (configurationApplied()),
				 this, SLOT (handleConfigurationApplied()));

	this->connect (rendererConfiguration, SIGNAL (reloadCGPrograms()),
				 this, SLOT (handleReloadCGPrograms()));

	this->connect (rendererConfiguration, SIGNAL (setBlendingEnabled(bool)),
				 this, SLOT (handleSetBlendingEnabled(bool)));

	backgroundColor = 0xFF8598b5;
}

HGLSplatRendererWidget::~HGLSplatRendererWidget() {
  cgDestroyProgram(cgVertexProgram);
  cgDestroyProgram(cgFragmentPassOne);
  cgDestroyProgram(cgFragmentPassTwo);
  cgDestroyProgram(cgFragmentNormalization);
  cgDestroyContext(cgContext);
}

void HGLSplatRendererWidget::setVisible (const bool enable) {
	visible = enable;
}

bool HGLSplatRendererWidget::isVisible() const {
	return visible;
}

QSize HGLSplatRendererWidget::sizeHint() const {
	return viewPortSize;
}

QSize HGLSplatRendererWidget::minimumSizeHint() const {
	return Configuration::getInstance()->getMinimumViewPortSize();
}

QSizePolicy HGLSplatRendererWidget::sizePolicy() const {
	return QSizePolicy (QSizePolicy::Preferred , QSizePolicy::Preferred);
}

void HGLSplatRendererWidget::setShadingEnabled (const bool enable) {
	shadingEnabled = enable;
}

bool HGLSplatRendererWidget::isShadingEnabled() const {
	return shadingEnabled;
}

void HGLSplatRendererWidget::setLightDirection (const Vector3D newLightDirection) {
	this->makeCurrent();
	
	lightDirection = newLightDirection;
}

Vector3D HGLSplatRendererWidget::getLightDirection() const {
	return lightDirection;
}

void HGLSplatRendererWidget::setBackgroundColor(QRgb newBackgroundColor) {
	this->makeCurrent();

	backgroundColor = newBackgroundColor;
	glClearColor(backgroundColor.red() / 255.0f, backgroundColor.green() / 255.0f, backgroundColor.blue() / 255.0f, 1.0);
}

QRgb HGLSplatRendererWidget::getBackgroundColor() const {
	return backgroundColor.rgb();
}

void HGLSplatRendererWidget::setSceneView (const MyDataTypes::TransformationMatrix16f newSceneViewMatrix) {
	memcpy (sceneViewMatrix, newSceneViewMatrix, sizeof (MyDataTypes::TransformationMatrix16f));
}

void HGLSplatRendererWidget::getSceneView (MyDataTypes::TransformationMatrix16f sceneViewMatrix) const {
	memcpy (sceneViewMatrix, this->sceneViewMatrix, sizeof (MyDataTypes::TransformationMatrix16f));
}

void HGLSplatRendererWidget::setViewFrustum (const MyDataTypes::ViewFrustum newViewFrustum) {
	
	this->makeCurrent();
	
	viewFrustum = newViewFrustum;

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity();
	gluPerspective (viewFrustum.fieldOfView, viewFrustum.aspectRatio,
		            viewFrustum.nearPlane,   viewFrustum.farPlane);
	glMatrixMode (GL_MODELVIEW);
}

MyDataTypes::ViewFrustum HGLSplatRendererWidget::getViewFrustum() const {
	return viewFrustum;
}

void HGLSplatRendererWidget::setViewPortSize (const QSize newViewPortSize) {

	// we do not want to update the configuration, it has already been updated
	// see resizeGL()
	updateViewPortSize = false;
	this->resize (newViewPortSize);
	// from now on we want to update the view port size in the Configuration again,
	// that is whenever this widget is resized by the user
	updateViewPortSize = true;
}

QSize HGLSplatRendererWidget::getViewPortSize() const {
	return this->size();
}

void HGLSplatRendererWidget::setSelectionVisualizationEnabled(const bool showSelection) {
}

bool HGLSplatRendererWidget::isSelectionVisualizationEnabled() const {
	return false;
}

void HGLSplatRendererWidget::setupDisplayLists() {

	int i;
	Scene* scene;
	Object* object;
	SurfelCollection* surfelCollection;
	SurfelInterface* surfel;
	Vector3D coordinate;
	QRgb color;
	float r, g, b;
	Vector3D axisU, axisV;
	float l1, l2;
	float spec, diff, amb, shininess;

	SurfelInterface *associatedSurfel;
	Vector3D normal, associatedNormal, intersectionDirection;

	scene = Scene::getInstance();

	for (object = scene->getFirstObject(), i = 0; object != 0; object = scene->getNextObject(), i++) {

		surfelCollection = object->getSurfelCollection();

		// display list with non-clipped surfels
		glNewList (displayLists.at(i), GL_COMPILE);
						  		
		glBegin (GL_POINTS);			
		for (surfel = surfelCollection->getFirstSurfel(); surfel != 0; surfel = surfelCollection->getNextSurfel()) {

			// non-clipped surfels
			if(!surfel->isFlagOn(SurfelInterface::INVISIBLE) && surfel->getAssociatedSurfel(0) == 0) {
				color = surfel->getDiffuseColor();
				r = (float)((color & 0x00ff0000) >> 16) / 255.f;
				g = (float)((color & 0x0000ff00) >> 8) / 255.f;
				b = (float)(color & 0x000000ff) / 255.f;
				glColor3f(r, g, b);

				// pass normalized tangent vectors in TEX0 and TEX1
				axisU = surfel->getAxisOne();
				l1 = axisU.getLength();
				axisU.normalize();
//axisU=Vector3D(0,1,0);
				glMultiTexCoord4fARB(GL_TEXTURE0_ARB, axisU[0], axisU[1], axisU[2], 0.f);
				axisV = surfel->getAxisTwo();
				l2 = axisV.getLength();
				axisV.normalize();
//axisV=Vector3D(1,0,0);
				glMultiTexCoord4fARB(GL_TEXTURE1_ARB, axisV[0], axisV[1], axisV[2], 0.f);

				// pass 1/(length^2) and length of tangent vectors in TEX2
				glMultiTexCoord4fARB(GL_TEXTURE2_ARB, 1/(l1*l1), 1/(l2*l2), l1, l2);

				// reflectance properties in TEX3
				spec = surfel->getSpecularCoefficient();
				diff = surfel->getDiffuseCoefficient();
				// factor 0.5 is to match lighting in the SurfelGPRenderer
				amb = 0.5f * surfel->getAmbientCoefficient();
				// we need to scale the shininess, since it will be clamped to [0..1]
				shininess = surfel->getShininess() / 128.f;
				glMultiTexCoord4fARB(GL_TEXTURE3_ARB, spec, diff, amb, shininess);

				// dummy intersection lines
				glMultiTexCoord4fARB(GL_TEXTURE4_ARB, -11.f, 10.f, -10.f, 10.f);
				glMultiTexCoord4fARB(GL_TEXTURE5_ARB, -11.f, 10.f, -10.f, 10.f);

				// point position
				coordinate = surfel->getPosition();
				glVertex4f (coordinate[0], coordinate[1], coordinate[2], 1.f);
			}
		}

		glEnd();
		glEndList();

		// display list with clipped surfels
		glNewList (clippedDisplayLists.at(i), GL_COMPILE);
						  		
		clippedDisplayListsEmpty.at(i) = true;

		glBegin (GL_POINTS);			
		for (surfel = surfelCollection->getFirstSurfel(); surfel != 0; surfel = surfelCollection->getNextSurfel()) {

			if(surfel->getAssociatedSurfel(0) != 0 && !surfel->isFlagOn(SurfelInterface::INVISIBLE)) {
				clippedDisplayListsEmpty.at(i) = false;

				color = surfel->getDiffuseColor();
				r = (float)((color & 0x00ff0000) >> 16) / 255.f;
				g = (float)((color & 0x0000ff00) >> 8) / 255.f;
				b = (float)(color & 0x000000ff) / 255.f;
				glColor3f(r, g, b);

				// pass normalized tangent vectors in TEX0 and TEX1
				axisU = surfel->getAxisOne();
				l1 = axisU.getLength();
				axisU.normalize();
//axisU=Vector3D(0,1,0);
				glMultiTexCoord4fARB(GL_TEXTURE0_ARB, axisU[0], axisU[1], axisU[2], 0.f);
				axisV = surfel->getAxisTwo();
				l2 = axisV.getLength();
				axisV.normalize();
//axisV=Vector3D(1,0,0);
				glMultiTexCoord4fARB(GL_TEXTURE1_ARB, axisV[0], axisV[1], axisV[2], 0.f);

				// pass 1/(length^2) and length of tangent vectors in TEX2
				glMultiTexCoord4fARB(GL_TEXTURE2_ARB, 1/(l1*l1), 1/(l2*l2), l1, l2);

				// reflectance properties in TEX3
				spec = surfel->getSpecularCoefficient();
				diff = surfel->getDiffuseCoefficient();
				// factor 0.5 is to match lighting in the SurfelGPRenderer
				amb = 0.5f * surfel->getAmbientCoefficient();
				// we need to scale the shininess, since it will be clamped to [0..1]
				shininess = surfel->getShininess() / 128.f;
				glMultiTexCoord4fARB(GL_TEXTURE3_ARB, spec, diff, amb, shininess);

				for(int aN=0; aN<2; aN++) {
					if((associatedSurfel = surfel->getAssociatedSurfel(aN)) != 0) {
						// determine intersection line of the surfel and its associated surfel
						Vector3D normal1 = surfel->getNormal();
						Vector3D normal2 = associatedSurfel->getNormal();
						Vector3D normal3 = Vector3D::crossProduct(normal1, normal2);

						// a point on the intersection line is computed as the intersection of the
						// tangent planes of the surfels and a third plane going through the average
						// of those two positions and being orghogonal to the other two planes (i.e., 
						// having the intersection direction as its normal)
						float d1 = Vector3D::dotProduct(normal1, surfel->getPosition());
						float d2 = Vector3D::dotProduct(normal2, associatedSurfel->getPosition());
						float d3 = Vector3D::dotProduct(normal3, 0.5 * (surfel->getPosition() + associatedSurfel->getPosition()));

						float temp1, temp2, temp3, temp4, temp5, temp6, temp7;
						Vector3D pos;
				
						temp1 = normal1[0]*normal3[1];   
						temp2 = normal1[0]*normal2[1];
						temp3 = normal3[0]*normal1[1];
						temp4 = normal2[1]*normal3[0];
						temp5 = normal3[1]*normal2[0];
						temp6 = normal2[0]*normal1[1];
						temp7 = 1/(-normal2[2]*temp1+normal3[2]*temp2+normal2[2]*temp3-normal1[2]*temp4+normal1[2]*temp5-normal3[2]*temp6);
						pos[0] = temp7*(-normal1[1]*normal3[2]*d2+normal1[1]*d3*normal2[2]-normal1[2]*normal2[1]*d3+normal1[2]*normal3[1]*d2-d1*normal3[1]*normal2[2]+d1*normal2[1]*normal3[2]);
						pos[1] = -temp7*(-normal1[0]*normal3[2]*d2+normal1[0]*d3*normal2[2]+normal3[0]*normal1[2]*d2+normal3[2]*normal2[0]*d1-d3*normal2[0]*normal1[2]-normal3[0]*d1*normal2[2]);
						pos[2] = temp7*(-d1*temp4+d3*temp2-d2*temp1+d1*temp5+d2*temp3-d3*temp6);

						// express two points on the intersection line in local coordinates of the splat plane
						float u0 = Vector3D::dotProduct(pos - surfel->getPosition(), axisU);
						float v0 = Vector3D::dotProduct(pos - surfel->getPosition(), axisV);

						pos = pos + normal3;
						float u1 = Vector3D::dotProduct(pos - surfel->getPosition(), axisU);
						float v1 = Vector3D::dotProduct(pos - surfel->getPosition(), axisV);

						// switch orientation of clip line if flag is set						
						if (surfel->isFlagOn (SurfelInterface::CLIP_ORIENTATION)) {
							float tmpU, tmpV;
							tmpU = u0;
							tmpV = v0;
							u0 = u1;
							v0 = v1;
							u1 = tmpU;
							v1 = tmpV;
						}

						// store intersection line in texture coordinates
						if(aN == 0) {
							glMultiTexCoord4fARB(GL_TEXTURE4_ARB, u0, v0, u1, v1);
						} else {
							glMultiTexCoord4fARB(GL_TEXTURE5_ARB, u0, v0, u1, v1);
						}
					} else {
						// dummy intersection line
						if(aN == 0) {
							glMultiTexCoord4fARB(GL_TEXTURE4_ARB, -11.f, 10.f, -10.f, 10.f);
						} else {
							glMultiTexCoord4fARB(GL_TEXTURE5_ARB, -11.f, 10.f, -10.f, 10.f);
						}
					}
				}

				// point position
				coordinate = surfel->getPosition();
				glVertex4f (coordinate[0], coordinate[1], coordinate[2], 1.f);
			}
		}

		glEnd();
		glEndList();

	}

	displayListsValid = true;
}

void HGLSplatRendererWidget::allocateDisplayLists() {
	uint nofObjects, i;
	GLuint listID;

	this->makeCurrent();		
	this->removeDisplayLists();

	nofObjects = Scene::getInstance()->getNofObjects();

	// allocate separate display lists for non-clipped and clipped splats
	displayLists.resize (nofObjects);
	clippedDisplayLists.resize (nofObjects);
	clippedDisplayListsEmpty.resize (nofObjects);

	listID = glGenLists (nofObjects);
	for (i = 0; i < nofObjects; i++) {
		displayLists.at(i) = listID + i;
	}

	listID = glGenLists (nofObjects);
	for (i = 0; i < nofObjects; i++) {
		clippedDisplayLists.at(i) = listID + i;
	}

	// allocated, but not yet created
	displayListsValid = false;
}

void HGLSplatRendererWidget::removeDisplayLists() {
  
	if (displayLists.count() > 0) {
		glDeleteLists (displayLists.at(0), displayLists.count());
	}
	displayListsValid = false;
}


// *****************
// protected methods
// *****************

void HGLSplatRendererWidget::initializeGL() {

	Vector3D lightDirection = Configuration::getInstance()->getLightDirection();
    setLightDirection(lightDirection);

	QRgb backgroundColor = Configuration::getInstance()->getBackgroundColor();
	setBackgroundColor(backgroundColor);

	glEnable (GL_CULL_FACE);
	glCullFace (GL_BACK);

	// we need to resize the display list ID array, since there might already
	// be Objects loaded already
	this->allocateDisplayLists();

	// this call also updates the display lists
	shadingEnabled = true;
    this->setShadingEnabled (shadingEnabled);
  
    glEnable(GL_DEPTH_TEST);
	glEnable(GL_ALPHA_TEST);

	// set default view frustum
	viewFrustum.fieldOfView = 90.f;
	viewFrustum.aspectRatio = 1.f;
	viewFrustum.nearPlane = 3.f;
	viewFrustum.farPlane = 5000.f;
	setViewFrustum(viewFrustum);

	// initialize gl extensions
	if(!glh_init_extensions(REQUIRED_EXTENSIONS)) {
		exit( -1 );
	}

	// set blend function to accumulate colors weigthed by alpha channel
	glBlendFuncSeparateEXT(GL_SRC_ALPHA, GL_ONE, GL_ONE, GL_ONE);

	depthOffset = rendererConfiguration->getDepthOffset();

	// initialize texture used for normalization
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_RECTANGLE_NV); 
	glGenTextures(1, &normalizationTexture);
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, normalizationTexture); 
	glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 
	glTexParameteri(GL_TEXTURE_RECTANGLE_NV, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);	
	glCopyTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGBA, 0, 0, viewPortSize.width(), viewPortSize.height(), 0);

	initializeCg();
	initializePointSprites();
	setUniformCGParameters();
}

void HGLSplatRendererWidget::resizeGL (int width, int height) {

	Configuration *configuration = Configuration::getInstance();
	if (updateViewPortSize == true && visible == true) {
		  configuration->setViewPortSize (QSize (width, height));
	}

	viewPortSize = QSize(width, height);
	glViewport (0, 0, viewPortSize.width(), viewPortSize.height());
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity();

	// adjust size of normalization texture
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glCopyTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGBA, 0, 0, viewPortSize.width(), viewPortSize.height(), 0);
	checkGLError("ResizeGL: Could not allocate texture.");

	// set scaling from normalized clip space to viewport. 
	// normalized clip space is [-1..1]x[-1..1].
	cgGLSetParameter2f(cgparam_vp_viewPortScaling, width/2, height/2);
}

void HGLSplatRendererWidget::paintGL () {

	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	if(!displayListsValid) {
		setupDisplayLists();
	}

	if(blendingEnabled) {

		// first pass
		cgGLBindProgram(cgFragmentPassOne);
		checkCgError("CG: Bind fragment program pass 1.");
		cgGLSetParameter1f(cgparam_vp_depthOffset, depthOffset);
		renderDisplayLists();

		// second pass
		cgGLBindProgram(cgFragmentPassTwo);
		checkCgError("CG: Bind fragment program pass 2.");
		glEnable(GL_BLEND);
		glDepthMask(GL_FALSE);
		cgGLSetParameter1f(cgparam_vp_depthOffset, 0);
		renderDisplayLists();

		glDisable(GL_BLEND);
		glDepthMask(GL_TRUE);

		// normalization pass
		glDisable(GL_VERTEX_PROGRAM_ARB);
		glDisable(GL_DEPTH_TEST);      

		// copy frame buffer to texture for normalization
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glCopyTexSubImage2D(GL_TEXTURE_RECTANGLE_NV, 0, 0, 0, 0, 0, viewPortSize.width(), viewPortSize.height());

		cgGLBindProgram(cgFragmentNormalization);
		checkCgError("CG: Bind fragment program normalization pass");

		// setup raster position
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		gluOrtho2D(0, viewPortSize.width()-1, 0, viewPortSize.height()-1);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		// draw textured quad
		glBegin( GL_QUADS );

		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0, 0);  
		glVertex2i(0, 0);

		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, viewPortSize.width(), 0);  
		glVertex2i(viewPortSize.width()-1, 0);

		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, viewPortSize.width(), viewPortSize.height());  
		glVertex2i(viewPortSize.width()-1, viewPortSize.height()-1);

		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0, viewPortSize.height());
		glVertex2i(0, viewPortSize.height()-1);

		glEnd();

		cgGLDisableTextureParameter(cgparam_fpn_normalizationTexture);

		// restore matrices
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glEnable(GL_VERTEX_PROGRAM_ARB);
		glEnable(GL_DEPTH_TEST);      

	} else {
		
		cgGLBindProgram(cgFragmentPassTwo);
		renderDisplayLists();
	}

	
	glLoadMatrixf (sceneViewMatrix);
	RendererInterface::callPostRenderCBs();
	
	glFinish();		

	emit repainted();
}

void HGLSplatRendererWidget::renderDisplayLists() {

	uint nofObjects, i;
	MyDataTypes::TransformationMatrix16f objectTransformation, finalTransformation, lightTrafo, inverseTrafo;
	Scene *scene;
	Object *object, *activeObject;
	Vector3D coordinate, normal;
	float cameraLightVector[4], objectLightVector[4];

	scene = Scene::getInstance();
	nofObjects = scene->getNofObjects();
	if (nofObjects == 0)
	  return;

	activeObject = scene->getActiveObject();
	for (object = scene->getFirstObject(), i = 0; object != 0; object = scene->getNextObject(), i++) {
		if (object->isFlagOn (Object::VISIBLE)) {

			object->getTransformationMatrix (objectTransformation);
			MtrMult4x4f (sceneViewMatrix, objectTransformation, finalTransformation);
			glLoadMatrixf (finalTransformation);

			cgGLSetStateMatrixParameter(cgparam_vp_ModelViewProjection,
								CG_GL_MODELVIEW_PROJECTION_MATRIX,
								CG_GL_MATRIX_IDENTITY);

			// transform light direction to object space
			MtrTranspose4x4f(finalTransformation, lightTrafo);
			cameraLightVector[0] = -lightDirection[0];
			cameraLightVector[1] = -lightDirection[1];
			cameraLightVector[2] = -lightDirection[2];
			cameraLightVector[3] = 0.f;
			MtrMultVector4f(lightTrafo, cameraLightVector, objectLightVector);
			cgGLSetParameter3f(cgparam_vp_lightDirection, objectLightVector[0], objectLightVector[1], objectLightVector[2]);

			// transform eye position (i.e., origin) to object space
			MtrInverse4x4f(finalTransformation, inverseTrafo);
			cgGLSetParameter3f(cgparam_vp_eyePosition, inverseTrafo[12], inverseTrafo[13], inverseTrafo[14]);

			// display list with non-clipped surfels
			glCallList (displayLists.at(i));

			// display list with clipped surfels
			if(clippedDisplayListsEmpty.at(i) == false) {
				glCallList (clippedDisplayLists.at(i));
			}
		}
	}
}

void HGLSplatRendererWidget::mousePressEvent (QMouseEvent *me) {	
	DefaultMouseHandler::handleMousePressEvent (me);
}

void HGLSplatRendererWidget::mouseMoveEvent (QMouseEvent *me) {
	DefaultMouseHandler::handleMouseMoveEvent (me);
}

void HGLSplatRendererWidget::mouseReleaseEvent (QMouseEvent *me) {
	DefaultMouseHandler::handleMouseReleaseEvent (me);
}

void HGLSplatRendererWidget::keyPressEvent (QKeyEvent *ke) {
	ke->ignore();
}

void HGLSplatRendererWidget::wheelEvent (QWheelEvent *we) {

	we->accept();
	if (we->state() & QWheelEvent::ControlButton) {
		Scene::getInstance()->getActiveObject()->translate (0.0f, 0.0f, we->delta());
	}
	else {
		Scene::getInstance()->translate (0.0f, 0.0f, we->delta());
	}
	
}

void HGLSplatRendererWidget::initializeCg() {

	cgContext = cgCreateContext();
	checkCgError();

	QString cgProgramDirectory = rendererConfiguration->getCgProgramDirectory();
	QString programFile;

	// vertex program
	programFile = cgProgramDirectory + "/vertexPassOne.cg";
	cgVertexProgram = cgCreateProgramFromFile(cgContext,
								CG_SOURCE, programFile,
								CG_PROFILE_VP30,
								"vertexPassOne", NULL);
	checkCgError();
	checkCgCompileError("CG: Compile vertex program.");

	cgGLLoadProgram(cgVertexProgram);
	checkCgError();

	// get parameters
	cgparam_vp_ModelViewProjection = cgGetNamedParameter(cgVertexProgram, "modelViewProjection");
	checkCgError();
	cgparam_vp_viewPortScaling = cgGetNamedParameter(cgVertexProgram, "viewPortScaling");
	checkCgError();
	cgparam_vp_F = cgGetNamedParameter(cgVertexProgram, "F");
	checkCgError();
	cgparam_vp_conditionThreshold = cgGetNamedParameter(cgVertexProgram, "conditionThreshold");
	checkCgError();
	cgparam_vp_depthOffset = cgGetNamedParameter(cgVertexProgram, "depthOffset");
	checkCgError();
	cgparam_vp_lightDirection = cgGetNamedParameter(cgVertexProgram, "lightDirection");
	checkCgError();
	cgparam_vp_eyePosition = cgGetNamedParameter(cgVertexProgram, "eyePosition");
	checkCgError();

	// fragment program pass one
	programFile = cgProgramDirectory + "/fragmentPassOne.cg";
	cgFragmentPassOne = cgCreateProgramFromFile(cgContext,
								CG_SOURCE, programFile.latin1(),
								CG_PROFILE_FP30,
								"fragmentPassOne", NULL);
	checkCgError();

	cgGLLoadProgram(cgFragmentPassOne);
	checkCgError();
	checkCgCompileError("CG: Compile fragment program pass one.");

	// get parameters
	cgparam_fp1_F = cgGetNamedParameter(cgFragmentPassOne, "F");


	// fragment program pass two
	programFile = cgProgramDirectory + "/fragmentPassTwo.cg";
	cgFragmentPassTwo = cgCreateProgramFromFile(cgContext,
								CG_SOURCE, programFile,
								CG_PROFILE_FP30,
								"fragmentPassTwo", NULL);
	checkCgError();

	cgGLLoadProgram(cgFragmentPassTwo);
	checkCgError();
	checkCgCompileError("CG: Compile fragment program pass two.");

	// get parameters
	cgparam_fp2_mytexture = cgGetNamedParameter(cgFragmentPassTwo, "mytexture");
	cgparam_fp2_F = cgGetNamedParameter(cgFragmentPassTwo, "F");


	// fragment program for normalization
	programFile = cgProgramDirectory + "/fragmentNormalization.cg";
	cgFragmentNormalization = cgCreateProgramFromFile(cgContext,
								CG_SOURCE, programFile,
								CG_PROFILE_FP30,
								"fragmentNormalization", NULL);
	checkCgError();

	cgGLLoadProgram(cgFragmentNormalization);
	checkCgError();
	checkCgCompileError("CG: Compile fragment program normalization pass.");

	// get parameters
	cgparam_fpn_normalizationTexture = cgGetNamedParameter(cgFragmentNormalization, "normalizationTexture");
	checkCgError();

	cgGLEnableProfile(CG_PROFILE_FP30);
	cgGLEnableProfile(CG_PROFILE_VP30);

	cgGLBindProgram(cgVertexProgram);
	checkCgError();
}

void HGLSplatRendererWidget::checkCgError(char *userMsg) {

	CGerror err = cgGetError();

	if (err != CG_NO_ERROR)
	{
		if(userMsg != 0) {
			qDebug("%s %s", userMsg, cgGetErrorString(err));
		} else {
			qDebug("%s", cgGetErrorString(err));
		}
	}
}

void HGLSplatRendererWidget::checkCgCompileError(char *userMsg) {

	if(cgGetLastListing(cgContext) != 0) {
		if(userMsg != 0) {
			qDebug("%s %s", userMsg, cgGetLastListing(cgContext));
		} else {
			qDebug("%s", cgGetLastListing(cgContext));
		}
	}
}

void HGLSplatRendererWidget::checkGLError(char *userMsg) {

	GLenum glError;

	glError = glGetError();
	if(glError != 0) {
		if(userMsg != 0) {
			qDebug("%s %i", userMsg, glError);
		} else {
			qDebug("OpenGL error: %i", glError);
		}
	}
}

void HGLSplatRendererWidget::initializePointSprites() {

	// enable point sprite rasterization
	glEnable(GL_POINT_SPRITE_NV);
	// enable point size specification by vertex program
	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE_ARB);

	// initialize point sprite texture
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glGenTextures(1, &pointSpriteTexture);
	glBindTexture(GL_TEXTURE_1D, pointSpriteTexture); 
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
}

void HGLSplatRendererWidget::setUniformCGParameters() {

	// get cutoff radius from configuration and pass it to vertex and fragment programs
	float r = rendererConfiguration->getCutOffRadius();
	r *= r;
	cgGLSetParameter1f(cgparam_vp_F, r);
	cgGLSetParameter1f(cgparam_fp1_F, r);
	cgGLSetParameter1f(cgparam_fp2_F, r);

	float ct = rendererConfiguration->getConditionThreshold();
	cgGLSetParameter1f(cgparam_vp_conditionThreshold, ct);

	cgGLSetParameter3f(cgparam_vp_lightDirection, lightDirection[0], lightDirection[1], lightDirection[2]);

	// generate Gauss texture for cutoff radius r
	float *tex = new float[gaussRes];
	genGaussTexture(r, gaussRes, tex);
	setGaussTexture(tex);
	delete[] tex;

	cgGLSetTextureParameter(cgparam_fp2_mytexture, pointSpriteTexture);
	cgGLEnableTextureParameter(cgparam_fp2_mytexture);

	// enable normalization texture for normalization fragment program
	cgGLSetTextureParameter(cgparam_fpn_normalizationTexture, normalizationTexture);
	cgGLEnableTextureParameter(cgparam_fpn_normalizationTexture);

	// set scaling from normalized clip space to viewport. 
	// normalized clip space is [-1..1]x[-1..1].
	cgGLSetParameter2f(cgparam_vp_viewPortScaling, viewPortSize.width()/2, viewPortSize.height()/2);
}

void HGLSplatRendererWidget::genGaussTexture(float r, int res, float *tex) {

	float x;
	int i;

	for (i=0; i<res; i++) {

// tent
//		tex[i] =  0.3f*(1-(float)i/(float)(res-1));

// quadratic spline
// the scale factor s is chosen making a trade-off between precision loss 
// and risk of framebuffer overflow
//		float u, s;
//		s = 0.4f;
//		u = 1.5f * (1 - (float)i/(float)(res-1));
//		if(u < 1) {
//			tex[i] = s * u*u * 0.5f; 
//		} else {
//			tex[i] = s * (-0.5f + u - (u-1.f)*(u-1.f));
//		}

// gaussian
		x = r*r * (float)i / ((float)res-1.f);
		tex[i] = 1/(2*M_PI) * exp(-0.5*x);
	}
}

void HGLSplatRendererWidget::setGaussTexture(float *tex) {
	
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glTexImage1D(GL_TEXTURE_1D, 0, GL_ALPHA, gaussRes, 0, GL_ALPHA, GL_FLOAT, tex);

	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
}

///////////////////////////////////////////////////////////////////////////////

// *************
// private slots
// *************

void HGLSplatRendererWidget::handleObjectModified () {
	displayListsValid = false;
}

void HGLSplatRendererWidget::handleActiveObjectChanged() {
	displayListsValid = false;
}

void HGLSplatRendererWidget::handleObjectAdded() {
	allocateDisplayLists();
}

void HGLSplatRendererWidget::handleObjectRemoved() {
	allocateDisplayLists();
}

void HGLSplatRendererWidget::handleCutOffRadiusChanged (const float newCutOffRadius) {

	float F = newCutOffRadius*newCutOffRadius;
	cgGLSetParameter1f(cgparam_vp_F, F);
	cgGLSetParameter1f(cgparam_fp1_F, F);
	cgGLSetParameter1f(cgparam_fp2_F, F);

	// generate Gauss texture for new cutoff radius
	float *tex = new float[gaussRes];
	genGaussTexture(newCutOffRadius, gaussRes, tex);
	setGaussTexture(tex);
	delete[] tex;
}

void HGLSplatRendererWidget::handleConditionThresholdChanged(const float newConditionThreshold) {
	cgGLSetParameter1f(cgparam_vp_conditionThreshold, newConditionThreshold);
}

void HGLSplatRendererWidget::handleDepthOffsetChanged(const float newDepthOffset) {
	depthOffset = newDepthOffset;
}


void HGLSplatRendererWidget::handleConfigurationApplied()
{
  this->update();
}

void HGLSplatRendererWidget::handleReloadCGPrograms() {
	cgDestroyProgram(cgVertexProgram);
	cgDestroyProgram(cgFragmentPassOne);
	cgDestroyProgram(cgFragmentPassTwo);
	cgDestroyProgram(cgFragmentNormalization);

	initializeCg();
	setUniformCGParameters();
}

void HGLSplatRendererWidget::handleSetBlendingEnabled(bool enable) {
	blendingEnabled = enable;
}


QImage HGLSplatRendererWidget::convertFrameBufferToQImage() {

	QImage image;
	image.create (viewPortSize, 32);
	int index;

	float *pixels = new float[viewPortSize.width() * viewPortSize.height() * 3];
	glReadPixels(0, 0, viewPortSize.width(), viewPortSize.height(), GL_RGB, GL_FLOAT, pixels);
	GLenum err = glGetError();
	switch(err) {
		case GL_INVALID_ENUM : qDebug("invalid enum");
		case GL_INVALID_OPERATION : qDebug("invalid operation");
	}

	for(int y=0; y<viewPortSize.height(); y++) {
		for(int x=0; x<viewPortSize.width(); x++) {
			index = viewPortSize.width()*y + x;
			*((uint *)image.scanLine(viewPortSize.height()-1-y) + x) = qRgb( (int)(255.f*pixels[index*3]), (int)(255.f*pixels[index*3+1]), (int)(255.f*pixels[index*3+2]) );
		}
	}

	delete[] pixels;

	return image;
}


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
