#ifndef _HGLSPLATRENDERERWIDGET_H_
#define _HGLSPLATRENDERERWIDGET_H_

#include <qgl.h>
#include <qevent.h>
#include <qarray.h>
#include <vector>

#include <Cg/cg.h>
#include <Cg/cgGL.h>

#include "../../../DataTypes/src/MyDataTypes.h"
#include "../../../Core/Interfaces/src/RendererInterface.h"
#include "HGLSplatRendererConfiguration.h"

/**
 * This class implements the HGLSplat preview renderer. The renderer is based on Jussi Rasanen's
 * splatting framework with perspectively correct splat shapes.
 *
 * @author Matthias Zwicker
 * @version 1.2
 */

#define REQUIRED_EXTENSIONS "GL_ARB_vertex_program " \
                            "GL_ARB_fragment_program " \
                            "GL_ARB_multitexture " \
							"GL_ARB_texture_cube_map " \
                            "GL_NV_point_sprite " \
	                        "GL_ARB_vertex_buffer_object " \
                            "GL_NV_texture_rectangle " \
                            "GL_EXT_blend_func_separate "


class HGLSplatRendererWidget : public QGLWidget
{
	Q_OBJECT

public:

	HGLSplatRendererWidget (const bool isVisible, HGLSplatRendererConfiguration *rendererConfiguration, QWidget *parent = 0, const char *name = 0);
	virtual ~HGLSplatRendererWidget();

	void setVisible (const bool enable);
	bool isVisible() const;

	QSize sizeHint() const;
	QSize minimumSizeHint() const;
	QSizePolicy sizePolicy() const;

	/**
	 * Enables/disables the lighting in HGLSplat. The
	 * image is <em>not</em> re-rendered automatically.
	 *
	 * @param enable
	 *        set to <code>true</code> if lighting should be enabled; <code>false</code> else
	 */
	void setShadingEnabled (const bool enable);

	/**
	 * Returns the status of the lighting in HGLSplat.
	 *
	 * @return <code>true</code> if lighting in HGLSplat is enabled; <code>false</code> else
	 */
	bool isShadingEnabled() const;

	void setLightDirection (const Vector3D newLightDirection);

	Vector3D getLightDirection() const;

	void setBackgroundColor(QRgb newBackgroundColor);
	QRgb getBackgroundColor() const;
	

	void setSceneView (const MyDataTypes::TransformationMatrix16f newSceneViewMatrix);
	void getSceneView (MyDataTypes::TransformationMatrix16f sceneViewMatrix) const;

	void setViewFrustum (const MyDataTypes::ViewFrustum newViewFrustum);
	MyDataTypes::ViewFrustum getViewFrustum() const;

	void setViewPortSize (const QSize newViewPortSize);
	QSize getViewPortSize() const;

	void setSelectionVisualizationEnabled(const bool showSelection);
	bool isSelectionVisualizationEnabled() const;

	QImage convertFrameBufferToQImage();

signals:

	/**
	 * Emitted whenever this <code>HGLSplatRendererWidget</code> has repainted its content.
	 *
	 * <b>Note:</b> Connected to the <code>HGLSplatRenderer</code> which on its turn emits
	 *              the signal <code>rendererWidgetRepainted</code>.
	 */
	void repainted();

protected:

	void mousePressEvent (QMouseEvent *me);
	void mouseMoveEvent (QMouseEvent *me);
	void mouseReleaseEvent (QMouseEvent *me);
	void keyPressEvent (QKeyEvent *ke);
	void wheelEvent (QWheelEvent *we);
	
	// inherited virtual methods from QGLWidget
	void initializeGL();
	void resizeGL (int width, int height);
	void paintGL();

private:

	bool displayListsValid;
	bool shadingEnabled;
	bool blendingEnabled;
	QSize viewPortSize;
	Vector3D lightDirection;
	QColor backgroundColor;
	MyDataTypes::TransformationMatrix16f sceneViewMatrix;
	MyDataTypes::viewFrustum viewFrustum;

	QArray<GLuint> displayLists;
	QArray<GLuint> clippedDisplayLists;
	QArray<bool> clippedDisplayListsEmpty;
	void setupDisplayLists();
	void allocateDisplayLists();
	void removeDisplayLists();

	HGLSplatRendererConfiguration *rendererConfiguration;
	bool visible, updateViewPortSize;

	CGcontext cgContext;

	CGprogram cgVertexProgram;
	CGprogram cgFragmentPassOne;
	CGprogram cgFragmentPassTwo;
	CGprogram cgFragmentNormalization;

	CGparameter cgparam_vp_ModelViewProjection;
	CGparameter cgparam_vp_viewPortScaling;
	CGparameter cgparam_vp_F;
	CGparameter cgparam_vp_conditionThreshold;
	CGparameter cgparam_vp_depthOffset;
	CGparameter cgparam_vp_lightDirection;
	CGparameter cgparam_vp_eyePosition;

	CGparameter cgparam_fp1_F;
	CGparameter cgparam_fp2_F;
	CGparameter cgparam_fp2_mytexture;
	CGparameter cgparam_fpn_normalizationTexture;

	void initializeCg();
	void initializePointSprites();
	void setUniformCGParameters();
	void genGaussTexture(float r, int res, float *tex);
	void setGaussTexture(float *tex);
	void renderDisplayLists();

	void checkCgError(char *userMsg = 0);
	void checkCgCompileError(char *userMsg = 0);
	void checkGLError(char *userMsg = 0);

	float depthOffset;
	const int gaussRes;
	GLuint pointSpriteTexture;

	GLuint normalizationTexture;

private slots:
	
	void handleObjectModified();
	void handleActiveObjectChanged();
	void handleObjectAdded();
	void handleObjectRemoved();
	void handleReloadCGPrograms();

	void handleCutOffRadiusChanged (const float newCutOffRadius);
	void handleConditionThresholdChanged (const float newConditionThreshold);
	void handleDepthOffsetChanged(const float newDepthOffset);
	void handleSetBlendingEnabled (bool enable);

	// connected to the HGLSplatRendererConfiguration, signal configurationApplied: re-renders the image
	void handleConfigurationApplied();  

};

#endif  // _HGLSPLATRENDERERWIDGET_H_

 /* Some Emacs-Hints -- please don't remove:
 *
 *  Local Variables:
 *  mode:C++
 *  tab-width:4
 *  comment-column:72
 *  End:
 */

